@@ -2,7 +2,7 @@
  * New Interface to Process Table -- PROCTAB Stream (a la Directory streams)
  * Copyright (C) 1996 Charles L. Blake.
  * Copyright (C) 1998 Michael K. Johnson
- * Copyright 1998-2002 Albert Cahalan
+ * Copyright 1998-2003 Albert Cahalan
  * May be distributed under the conditions of the
  * GNU Library General Public License; a copy is in COPYING
  */
@@ -31,6 +31,7 @@
 #include <fs_secure.h>
 #endif
 
+// sometimes it's easier to do this manually, w/o gcc helping
 #ifdef PROF
 extern void __cyg_profile_func_enter(void*,void*);
 #define ENTER(x) __cyg_profile_func_enter((void*)x,(void*)x)
@@ -40,51 +41,7 @@ extern void __cyg_profile_func_enter(void*,void*);
 #define LEAVE(x)
 #endif
 
-/* initiate a process table scan
- */
-PROCTAB* openproc(int flags, ...) {
-    va_list ap;
-    PROCTAB* PT = xmalloc(sizeof(PROCTAB));
-    
-    if (flags & PROC_PID)
-      PT->procfs = NULL;
-    else if (!(PT->procfs = opendir("/proc")))
-      return NULL;
-    PT->flags = flags;
-    va_start(ap, flags);		/*  Init args list */
-    if (flags & PROC_PID)
-    	PT->pids = va_arg(ap, pid_t*);
-    else if (flags & PROC_UID) {
-    	PT->uids = va_arg(ap, uid_t*);
-	PT->nuid = va_arg(ap, int);
-    }
-    va_end(ap);				/*  Clean up args list */
-    return PT;
-}
-
-/* terminate a process table scan
- */
-void closeproc(PROCTAB* PT) {
-    if (PT){
-        if (PT->procfs) closedir(PT->procfs);
-        free(PT);
-    }
-}
-
-/* deallocate the space allocated by readproc if the passed rbuf was NULL
- */
-void freeproc(proc_t* p) {
-    if (!p)	/* in case p is NULL */
-	return;
-    /* ptrs are after strings to avoid copying memory when building them. */
-    /* so free is called on the address of the address of strvec[0]. */
-    if (p->cmdline)
-	free((void*)*p->cmdline);
-    if (p->environ)
-	free((void*)*p->environ);
-    free(p);
-}
-
+static int task_dir_missing;
 
 ///////////////////////////////////////////////////////////////////////////
 
@@ -256,7 +213,7 @@ ENTER(0x220);
         P->ppid = strtol(S,&S,10);
         continue;
     case_Pid:
-        P->pid = strtol(S,&S,10);
+        P->tid = strtol(S,&S,10);
         continue;
 
     case_ShdPnd:
@@ -505,65 +462,24 @@ int read_cmdline(char *restrict const dst, unsigned sz, unsigned pid){
 	    i < n && l[i] == x;			\
 	} )
 
-/* readproc: return a pointer to a proc_t filled with requested info about the
- * next process available matching the restriction set.  If no more such
- * processes are available, return a null pointer (boolean false).  Use the
- * passed buffer instead of allocating space if it is non-NULL.  */
+//////////////////////////////////////////////////////////////////////////////////
+// This reads process info from /proc in the traditional way, for one process.
+// The pid (tgid? tid?) is already in p, and a path to it in path, with some
+// room to spare.
+static proc_t* simple_readproc(PROCTAB *restrict const PT, proc_t *restrict const p) {
+    static struct stat sb;		// stat() buffer
+    static char sbuf[1024];	// buffer for stat,statm
+    char *restrict const path = PT->path;
+    unsigned flags = PT->flags;
 
-/* This is optimized so that if a PID list is given, only those files are
- * searched for in /proc.  If other lists are given in addition to the PID list,
- * the same logic can follow through as for the no-PID list case.  This is
- * fairly complex, but it does try to not to do any unnecessary work.
- */
-proc_t* readproc(PROCTAB* PT, proc_t* p) {
-    static struct direct *ent;		/* dirent handle */
-    static struct stat sb;		/* stat buffer */
-    static char path[32], sbuf[1024];	/* bufs for stat,statm */
-#ifdef FLASK_LINUX
-    security_id_t secsid;
-#endif
-    pid_t pid;  // saved until we have a proc_t allocated for sure
-
-    /* loop until a proc matching restrictions is found or no more processes */
-    /* I know this could be a while loop -- this way is easier to indent ;-) */
-next_proc:				/* get next PID for consideration */
-
-/*printf("PT->flags is 0x%08x\n", PT->flags);*/
-#define flags (PT->flags)
-
-    if (flags & PROC_PID) {
-        pid = *(PT->pids)++;
-	if (unlikely(!pid)) return NULL;
-	snprintf(path, sizeof path, "/proc/%d", pid);
-    } else {					/* get next numeric /proc ent */
-	for (;;) {
-	    ent = readdir(PT->procfs);
-	    if(unlikely(unlikely(!ent) || unlikely(!ent->d_name))) return NULL;
-	    if(likely( likely(*ent->d_name > '0') && likely(*ent->d_name <= '9') )) break;
-	}
-	pid = strtoul(ent->d_name, NULL, 10);
-	memcpy(path, "/proc/", 6);
-	strcpy(path+6, ent->d_name);  // trust /proc to not contain evil top-level entries
-//	snprintf(path, sizeof path, "/proc/%s", ent->d_name);
-    }
-#ifdef FLASK_LINUX
-    if ( stat_secure(path, &sb, &secsid) == -1 ) /* no such dirent (anymore) */
-#else
     if (unlikely(stat(path, &sb) == -1))	/* no such dirent (anymore) */
-#endif
 	goto next_proc;
 
     if ((flags & PROC_UID) && !XinLN(uid_t, sb.st_uid, PT->uids, PT->nuid))
 	goto next_proc;			/* not one of the requested uids */
 
-    if (!p)
-	p = xcalloc(p, sizeof *p); /* passed buf or alloced mem */
-
     p->euid = sb.st_uid;			/* need a way to get real uid */
-#ifdef FLASK_LINUX
-    p->secsid = secsid;
-#endif
-    p->pid  = pid;
+    p->egid = sb.st_gid;			/* need a way to get real gid */
 
     if (flags & PROC_FILLSTAT) {         /* read, parse /proc/#/stat */
 	if (unlikely( file2str(path, "stat", sbuf, sizeof sbuf) == -1 ))
@@ -571,7 +487,7 @@ next_proc:				/* get next PID for consideration */
 	stat2proc(sbuf, p);				/* parse /proc/#/stat */
     }
 
-    if (unlikely(flags & PROC_FILLMEM)) {				/* read, parse /proc/#/statm */
+    if (unlikely(flags & PROC_FILLMEM)) {	/* read, parse /proc/#/statm */
 	if (likely( file2str(path, "statm", sbuf, sizeof sbuf) != -1 ))
 	    statm2proc(sbuf, p);		/* ignore statm errors here */
     }						/* statm fields just zero */
@@ -613,10 +529,197 @@ next_proc:				/* get next PID for consideration */
         p->environ = NULL;
     
     return p;
+next_proc:
+    return NULL;
 }
-#undef flags
 
-/* ps_readproc: return a pointer to a proc_t filled with requested info about the
+//////////////////////////////////////////////////////////////////////////////////
+// This reads /proc/*/task/* data, for one task.
+// p is the POSIX process (task group summary) (not needed by THIS implementation)
+// t is the POSIX thread (task group member, generally not the leader)
+// path is a path to the task, with some room to spare.
+static proc_t* simple_readtask(PROCTAB *restrict const PT, const proc_t *restrict const p, proc_t *restrict const t, char *restrict const path) {
+    static struct stat sb;		// stat() buffer
+    static char sbuf[1024];	// buffer for stat,statm
+    unsigned flags = PT->flags;
+
+    if (unlikely(stat(path, &sb) == -1))	/* no such dirent (anymore) */
+	goto next_task;
+
+//    if ((flags & PROC_UID) && !XinLN(uid_t, sb.st_uid, PT->uids, PT->nuid))
+//	goto next_task;			/* not one of the requested uids */
+
+    t->euid = sb.st_uid;			/* need a way to get real uid */
+    t->egid = sb.st_gid;			/* need a way to get real gid */
+
+    if (flags & PROC_FILLSTAT) {         /* read, parse /proc/#/stat */
+	if (unlikely( file2str(path, "stat", sbuf, sizeof sbuf) == -1 ))
+	    goto next_task;			/* error reading /proc/#/stat */
+	stat2proc(sbuf, t);				/* parse /proc/#/stat */
+    }
+
+    if (unlikely(flags & PROC_FILLMEM)) {	/* read, parse /proc/#/statm */
+	if (likely( file2str(path, "statm", sbuf, sizeof sbuf) != -1 ))
+	    statm2proc(sbuf, t);		/* ignore statm errors here */
+    }						/* statm fields just zero */
+
+    if (flags & PROC_FILLSTATUS) {         /* read, parse /proc/#/status */
+       if (likely( file2str(path, "status", sbuf, sizeof sbuf) != -1 )){
+           status2proc(sbuf, t);
+       }
+    }
+
+    /* some number->text resolving which is time consuming */
+    if (flags & PROC_FILLUSR){
+	strncpy(t->euser,   user_from_uid(t->euid), sizeof t->euser);
+        if(flags & PROC_FILLSTATUS) {
+            strncpy(t->ruser,   user_from_uid(t->ruid), sizeof t->ruser);
+            strncpy(t->suser,   user_from_uid(t->suid), sizeof t->suser);
+            strncpy(t->fuser,   user_from_uid(t->fuid), sizeof t->fuser);
+        }
+    }
+
+    /* some number->text resolving which is time consuming */
+    if (flags & PROC_FILLGRP){
+        strncpy(t->egroup, group_from_gid(t->egid), sizeof t->egroup);
+        if(flags & PROC_FILLSTATUS) {
+            strncpy(t->rgroup, group_from_gid(t->rgid), sizeof t->rgroup);
+            strncpy(t->sgroup, group_from_gid(t->sgid), sizeof t->sgroup);
+            strncpy(t->fgroup, group_from_gid(t->fgid), sizeof t->fgroup);
+        }
+    }
+
+#if 0
+    if ((flags & PROC_FILLCOM) || (flags & PROC_FILLARG))	/* read+parse /proc/#/cmdline */
+	t->cmdline = file2strvec(path, "cmdline");
+    else
+        t->cmdline = NULL;
+
+    if (unlikely(flags & PROC_FILLENV))			/* read+parse /proc/#/environ */
+	t->environ = file2strvec(path, "environ");
+    else
+        t->environ = NULL;
+#else
+    t->cmdline = p->cmdline;  // better not free these until done with all threads!
+    t->environ = p->environ;
+#endif
+
+    return t;
+next_task:
+    return NULL;
+}
+
+//////////////////////////////////////////////////////////////////////////////////
+// This finds processes in /proc in the traditional way.
+// Return non-zero on success.
+static int simple_nextpid(PROCTAB *restrict const PT, proc_t *restrict const p) {
+  static struct direct *ent;		/* dirent handle */
+  char *restrict const path = PT->path;
+  for (;;) {
+    ent = readdir(PT->procfs);
+    if(unlikely(unlikely(!ent) || unlikely(!ent->d_name))) return 0;
+    if(likely( likely(*ent->d_name > '0') && likely(*ent->d_name <= '9') )) break;
+  }
+  p->tgid = strtoul(ent->d_name, NULL, 10);
+  p->tid = p->tgid;
+  memcpy(path, "/proc/", 6);
+  strcpy(path+6, ent->d_name);  // trust /proc to not contain evil top-level entries
+  return 1;
+}
+
+//////////////////////////////////////////////////////////////////////////////////
+// This finds tasks in /proc/*/task/ in the traditional way.
+// Return non-zero on success.
+static int simple_nexttid(PROCTAB *restrict const PT, const proc_t *restrict const p, proc_t *restrict const t, char *restrict const path) {
+  static struct direct *ent;		/* dirent handle */
+  (void)p;
+  if(!PT->taskdir){
+    // use "path" as some tmp space
+    snprintf(path, PROCPATHLEN, "%s/task", PT->path);
+    PT->taskdir = opendir(path);
+    if(!PT->taskdir) return 0;
+  }
+  for (;;) {
+    ent = readdir(PT->taskdir);
+    if(unlikely(unlikely(!ent) || unlikely(!ent->d_name))) return 0;
+    if(likely( likely(*ent->d_name > '0') && likely(*ent->d_name <= '9') )) break;
+  }
+  t->tid = strtoul(ent->d_name, NULL, 10);
+  t->tgid = p->tgid;
+  t->ppid = p->ppid;  // cover for kernel behavior? we want both actually...?
+  snprintf(path, PROCPATHLEN, "%s/task/%s", PT->path, ent->d_name);
+  return 1;
+}
+
+//////////////////////////////////////////////////////////////////////////////////
+// This "finds" processes in a list that was given to openproc().
+// Return non-zero on success. (tgid was handy)
+static int listed_nextpid(PROCTAB *restrict const PT, proc_t *restrict const p) {
+  char *restrict const path = PT->path;
+  pid_t tgid = *(PT->pids)++;
+  if(likely( tgid )){
+    snprintf(path, PROCPATHLEN, "/proc/%d", tgid);
+    p->tgid = tgid;
+    p->tid = tgid;  // they match for leaders
+  }
+  return tgid;
+}
+
+//////////////////////////////////////////////////////////////////////////////////
+// This "finds" processes by guessing every possible one of them!
+// Return non-zero on success. (pid was handy)
+#if 0
+static int stupid_nextpid(PROCTAB *restrict const PT, proc_t *restrict const p) {
+  char *restrict const path = PT->path;
+  pid_t pid = --PT->u;
+  if(likely( pid )){
+    snprintf(path, PROCPATHLEN, "/proc/%d", pid);
+    p->pid = pid;
+  }
+  return pid;
+}
+#endif
+
+//////////////////////////////////////////////////////////////////////////////////
+// This reads process info from proc_t structs already attached to a PROCTAB.
+// Yeah, we don't retain any pointer for freeing the memory later. Oh well.
+// This code is for development only.
+#if 0
+static proc_t* predone_readproc(PROCTAB *restrict const PT, proc_t *restrict const p) {
+  proc_t *tmp;
+  proc_t *ret = NULL;
+  for(;;){
+    tmp = PT->vp;
+    if(!tmp) _exit(49);  // can't happen
+    PT->vp = tmp->next;
+    if(tmp->pid == tmp->tgid){   // got a leader?
+      memcpy(p,tmp,sizeof(proc_t));  // copy it, pointers and all
+      ret = p;
+      break;
+    }
+  }
+  if(!ret) _exit(99); // can't happen
+  while(PT->vp){
+    tmp = PT->vp;
+    if(tmp->pid == tmp->tgid) break;  // OK, next one is a leader
+    PT->vp = tmp->next;
+  }
+  return ret;
+}
+#endif
+
+//////////////////////////////////////////////////////////////////////////////////
+// This "finds" processes by pulling them off of a list.
+// Return non-zero on success.
+#if 0
+static int predone_nextpid(PROCTAB *restrict const PT, proc_t *restrict const p) {
+  (void)p;
+  return !!PT->vp;
+}
+#endif
+
+//////////////////////////////////////////////////////////////////////////////////
+/* readproc: return a pointer to a proc_t filled with requested info about the
  * next process available matching the restriction set.  If no more such
  * processes are available, return a null pointer (boolean false).  Use the
  * passed buffer instead of allocating space if it is non-NULL.  */
@@ -626,98 +729,179 @@ next_proc:				/* get next PID for consideration */
  * the same logic can follow through as for the no-PID list case.  This is
  * fairly complex, but it does try to not to do any unnecessary work.
  */
-proc_t* ps_readproc(PROCTAB* PT, proc_t* p) {
-    static struct direct *ent;		/* dirent handle */
-    static struct stat sb;		/* stat buffer */
-    static char path[32], sbuf[1024];	/* bufs for stat,statm */
-#ifdef FLASK_LINUX
-    security_id_t secsid;
-#endif
-    pid_t pid;  // saved until we have a proc_t allocated for sure
+proc_t* readproc(PROCTAB *restrict const PT, proc_t *restrict p) {
+  proc_t *ret;
+  proc_t *saved_p;
 
-    /* loop until a proc matching restrictions is found or no more processes */
-    /* I know this could be a while loop -- this way is easier to indent ;-) */
-next_proc:				/* get next PID for consideration */
+  if (PT->did_fake) PT->did_fake=0;
+  if (PT->taskdir) {
+    closedir(PT->taskdir);
+    PT->taskdir = NULL;
+  }
 
-/*printf("PT->flags is 0x%08x\n", PT->flags);*/
-#define flags (PT->flags)
+  saved_p = p;
+  if(!p) p = xcalloc(p, sizeof *p); /* passed buf or alloced mem */
 
-    for (;;) {
-	ent = readdir(PT->procfs);
-	if(unlikely(unlikely(!ent) || unlikely(!ent->d_name))) return NULL;
-	if(likely( likely(*ent->d_name > '0') && likely(*ent->d_name <= '9') )) break;
-    }
-    pid = strtoul(ent->d_name, NULL, 10);
-    memcpy(path, "/proc/", 6);
-    strcpy(path+6, ent->d_name);  // trust /proc to not contain evil top-level entries
-//  snprintf(path, sizeof path, "/proc/%s", ent->d_name);
+  for(;;){
+    // fills in the path, plus p->tid and p->tgid
+    if (unlikely(! PT->finder(PT,p) )) goto out;
 
-#ifdef FLASK_LINUX
-    if (stat_secure(path, &sb, &secsid) == -1) /* no such dirent (anymore) */
-#else
-    if (stat(path, &sb) == -1)		/* no such dirent (anymore) */
-#endif
-	goto next_proc;
+    // go read the process data
+    ret = PT->reader(PT,p);
+    if(ret) return ret;
+  }
 
-    if (!p)
-	p = xcalloc(p, sizeof *p); /* passed buf or alloced mem */
+out:
+  if(!saved_p) free(p);
+  // FIXME: maybe set tid to -1 here, for "-" in display?
+  return NULL;
+}
 
-    p->euid = sb.st_uid;			/* need a way to get real uid */
-#ifdef FLASK_LINUX
-    p->secsid = secsid;
+//////////////////////////////////////////////////////////////////////////////////
+// readtask: return a pointer to a proc_t filled with requested info about the
+// next task available.  If no more such tasks are available, return a null
+// pointer (boolean false).  Use the passed buffer instead of allocating
+// space if it is non-NULL.
+proc_t* readtask(PROCTAB *restrict const PT, const proc_t *restrict const p, proc_t *restrict t) {
+  static char path[PROCPATHLEN];       // must hold /proc/2000222000/task/2000222000/cmdline
+  proc_t *ret;
+  proc_t *saved_t;
+
+  saved_t = t;
+  if(!t) t = xcalloc(t, sizeof *t); /* passed buf or alloced mem */
+
+  if(task_dir_missing){  // got to fake a thread for old kernels
+    if(PT->did_fake) goto out;
+    PT->did_fake=1;
+    memcpy(t,p,sizeof(proc_t));
+    return t;
+  }
+
+  for(;;){
+    // fills in the path, plus t->tid and t->tgid
+    if (unlikely(! PT->taskfinder(PT,p,t,path) )) goto out;  // simple_nexttid
+
+    // go read the task data
+    ret = PT->taskreader(PT,p,t,path);          // simple_readtask
+    if(ret) return ret;
+  }
+
+out:
+  if(!saved_t) free(t);
+  return NULL;
+}
+
+//////////////////////////////////////////////////////////////////////////////////
+
+#if 0
+static void evil_grouping_hack(PROCTAB* PT){
+  proc_t *tp;
+  // first we read them
+  for(;;){
+    tp = malloc(sizeof(proc_t));
+    if(!tp) _exit(2);
+    if(!readproc(PT, tp)){
+      free(tp);
+      break;
+    }
+    tp->ring = tp;
+    tp->next = PT->vp;
+    PT->vp   = tp;
+  }
+  // now we scan
+  tp = PT->vp;
+  while(tp){
+    if(tp->pid == tp->tgid){  // if we found a leader
+      proc_t *tmp = PT->vp;
+      while(tmp){
+        if(tmp != tp && tmp->tgid == tp->tgid){
+          tmp->ring = tp->ring;
+          tp->ring = tmp;
+        }
+        tmp = tmp->next;
+      }
+    }
+    tp = tp->next;
+  }
+  // later, readproc returns what we already have
+  PT->finder = predone_nextpid;
+  PT->reader = predone_readproc;
+}
 #endif
-    p->pid  = pid;
-
-    if ((file2str(path, "stat", sbuf, sizeof sbuf)) == -1)
-	goto next_proc;			/* error reading /proc/#/stat */
-    stat2proc(sbuf, p);				/* parse /proc/#/stat */
-
-    if (flags & PROC_FILLMEM) {				/* read, parse /proc/#/statm */
-	if ((file2str(path, "statm", sbuf, sizeof sbuf)) != -1 )
-	    statm2proc(sbuf, p);		/* ignore statm errors here */
-    }						/* statm fields just zero */
-
-  /*  if (flags & PROC_FILLSTATUS) { */        /* read, parse /proc/#/status */
-       if ((file2str(path, "status", sbuf, sizeof sbuf)) != -1 ){
-           status2proc(sbuf, p);
-       }
-/*    }*/
-
-    /* some number->text resolving which is time consuming */
-    if (flags & PROC_FILLUSR){
-	strncpy(p->euser,   user_from_uid(p->euid), sizeof p->euser);
-/*        if(flags & PROC_FILLSTATUS) { */
-            strncpy(p->ruser,   user_from_uid(p->ruid), sizeof p->ruser);
-            strncpy(p->suser,   user_from_uid(p->suid), sizeof p->suser);
-            strncpy(p->fuser,   user_from_uid(p->fuid), sizeof p->fuser);
-/*        }*/
+
+// initiate a process table scan
+PROCTAB* openproc(int flags, ...) {
+    va_list ap;
+    struct stat sbuf;
+    static int did_stat;
+    PROCTAB* PT = xmalloc(sizeof(PROCTAB));
+
+    if(!did_stat){
+      task_dir_missing = stat("/proc/self/task", &sbuf);
+      did_stat = 1;
     }
+    PT->taskdir = NULL;
+    PT->taskfinder = simple_nexttid;
+    PT->taskreader = simple_readtask;
+
+    PT->reader = simple_readproc;
+    if (flags & PROC_PID){
+      PT->procfs = NULL;
+      PT->finder = listed_nextpid;
+    }else{
+      PT->procfs = opendir("/proc");
+      if(!PT->procfs) return NULL;
+      PT->finder = simple_nextpid;
+    }
+    PT->flags = flags;
+
+#if 0
+    if(getenv("EVIL_FINDER_HACK")){  // for development only
+      PT->finder = stupid_nextpid;
+      PT->u = 10000;
+    }
+#endif
+
+    va_start(ap, flags);		/*  Init args list */
+    if (flags & PROC_PID)
+    	PT->pids = va_arg(ap, pid_t*);
+    else if (flags & PROC_UID) {
+    	PT->uids = va_arg(ap, uid_t*);
+	PT->nuid = va_arg(ap, int);
+    }
+    va_end(ap);				/*  Clean up args list */
+
+#if 0
+    if(getenv("EVIL_GROUPING_HACK")) evil_grouping_hack(PT);
+#endif
+
+    return PT;
+}
 
-    /* some number->text resolving which is time consuming */
-    if (flags & PROC_FILLGRP){
-        strncpy(p->egroup, group_from_gid(p->egid), sizeof p->egroup);
-/*        if(flags & PROC_FILLSTATUS) { */
-            strncpy(p->rgroup, group_from_gid(p->rgid), sizeof p->rgroup);
-            strncpy(p->sgroup, group_from_gid(p->sgid), sizeof p->sgroup);
-            strncpy(p->fgroup, group_from_gid(p->fgid), sizeof p->fgroup);
-/*        }*/
+// terminate a process table scan
+void closeproc(PROCTAB* PT) {
+    if (PT){
+        if (PT->procfs) closedir(PT->procfs);
+        if (PT->taskdir) closedir(PT->taskdir);
+        free(PT);
     }
+}
 
-    if ((flags & PROC_FILLCOM) || (flags & PROC_FILLARG))	/* read+parse /proc/#/cmdline */
-	p->cmdline = file2strvec(path, "cmdline");
-    else
-        p->cmdline = NULL;
-
-    if (flags & PROC_FILLENV)			/* read+parse /proc/#/environ */
-	p->environ = file2strvec(path, "environ");
-    else
-        p->environ = NULL;
-    
-    return p;
+// deallocate the space allocated by readproc if the passed rbuf was NULL
+void freeproc(proc_t* p) {
+    if (!p)	/* in case p is NULL */
+	return;
+    /* ptrs are after strings to avoid copying memory when building them. */
+    /* so free is called on the address of the address of strvec[0]. */
+    if (p->cmdline)
+	free((void*)*p->cmdline);
+    if (p->environ)
+	free((void*)*p->environ);
+    free(p);
 }
-#undef flags
 
 
+//////////////////////////////////////////////////////////////////////////////////
 void look_up_our_self(proc_t *p) {
     char sbuf[1024];
 
@@ -732,8 +916,8 @@ HIDDEN_ALIAS(readproc);
 
 /* Convenient wrapper around openproc and readproc to slurp in the whole process
  * table subset satisfying the constraints of flags and the optional PID list.
- * Free allocated memory with freeproctab().  Access via tab[N]->member.  The
- * pointer list is NULL terminated.
+ * Free allocated memory with exit().  Access via tab[N]->member.  The pointer
+ * list is NULL terminated.
  */
 proc_t** readproctab(int flags, ...) {
     PROCTAB* PT = NULL;
